/*-------------------------------------------------------------------*/
/*                                                                   */
/* SAGAIS.C      Juha Kivijrvi  1998-2005                           */
/*               Olli Virmajoki  2005                                */
/*                                                                   */
/* - Adaptive genetic algorithm for VQ                               */
/* - SAGA with IS                                                    */
/*                                                                   */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <values.h>
#include <limits.h>

#if defined(__MSDOS__)
#include <conio.h>
#define AUMLAUT '\x84'
#else 
#define AUMLAUT '\xE4'
#endif

#define  FACTFILE "sagais.fac"

#include "memctrl.h"
#include "parametr.c"
#include "random.h"
#include "file.h"
#include "cb.h"
#include "owntypes.h"
#include "pnn.h"
#include "solution.h"

/*--------------------------------------------------------------------*/

#define  ProgName       "SAGAIS"
#define  VersionNumber  "version 0.02"    /*                        */
#define  LastUpdated    "24.1.05 "        /*  (by   Juha Kivijrvi) */

/*---------------------  Constants  ----------------------------------*/

#define CrossingMethodCount  7
#define EPSILON              0.0000001

/*---------------------  Macros  -------------------------------------*/

#define errorf ( Value(ErrorFunction) == MSE_D     ? MSE     : \
               ( Value(ErrorFunction) == SC_D      ? SC      : \
               ( Value(ErrorFunction) == JACCARD_D ? JACCARD : \
               ( Value(ErrorFunction) == DICE_D    ? DICE    : MSE ) ) ) )

#define LimitN(N, min, max)  ( (N)<(min) ? (min) : ((N)>(max) ? (max) : (N)) )
#define  round(a)		((llong) (a+0.5))
void RemoveEmptyPartitions(TRAININGSET* TS,
                           SOLUTION*    S);

/*---------------------  Types  --------------------------------------*/


typedef struct { int CrossingMethod;
                 int MutationProb;
                 int MaxNoise;
                 double Error;      } GENPARAMETERS;

typedef struct { SOLUTION      S;
                 GENPARAMETERS Parameters; } INDIVIDUAL;

typedef struct { clock_t latestclock;
                 int     carry;       } CLOCKINFO;


/*=============  I N D I V I D U A L    D E F I N I T I O N S  ======*/


#define IGenes(I)           ( I->S )
#define ICrossingMethod(I)  ( I->Parameters.CrossingMethod )
#define IMutationProb(I)    ( I->Parameters.MutationProb )
#define IMaxNoise(I)        ( I->Parameters.MaxNoise )
#define IError(I)           ( I->Parameters.Error )

                        
/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
printf("%s     %s                (last modified %s)\n",ProgName,
                                             VersionNumber,LastUpdated);
printf("This program creates random codebooks and tries to find\n");
printf("better ones with an adaptive genetic algorithm.\n\n");
printf("        Usage: SAGAIS [options] TSfile outputfile \n\n");
printf("        options:\n");
PrintOptions();
printf("%s     %s          Juha Kivij%crvi 1998-2005,  Olli Virmajoki 2005\n",
                                               ProgName,VersionNumber,AUMLAUT);
}


/*-------------------------------------------------------------------*/


void PrintElapsedTime(CLOCKINFO* ci)
{
llong time_elapsed;
long h,m,s;

time_elapsed = ( (llong) clock() + ci->carry * (llong) UINT_MAX ) /
                  CLOCKS_PER_SEC;

h = time_elapsed / 3600;
m = (time_elapsed % 3600) / 60;
s = time_elapsed % 60;

printf("%ld:%02ld:%02ld",h, m, s);
}        


/*-------------------------------------------------------------------*/


void PrintTime(CLOCKINFO* ci)
{
/* long h, m, s;
long processortime; */

printf("Time elapsed ");
PrintElapsedTime(ci);
printf("\n");

/*
processortime = (long)(clock()/CLOCKS_PER_SEC);

if (processortime>0)
    {
    h = processortime/3600;
    m = (processortime % 3600)/60;
    s = processortime%60;

    printf("Processor time used %ld:%02ld:%02ld\n", h, m, s);
    }
*/
}


/*-------------------------------------------------------------------*/


void UpdateClock(CLOCKINFO* ci)
{
clock_t cl;

cl = clock();
if ( cl < ci->latestclock ) ci->carry++;
ci->latestclock = cl;
}


/*-------------------------------------------------------------------*/


void PrintProgress(TRAININGSET* TS,
                   INDIVIDUAL*  Is[],
                   int          Generation,
                   CLOCKINFO*   ci)
{
int i;

if ( Value(QuietLevel) > 0 )
    {
    printf("%3d: %11.6f   c=%d  m=%5.2f%%  n=%3d   ", Generation,
           PrintableError(IError(Is[0]), TS), ICrossingMethod(Is[0]),
           IMutationProb(Is[0]) / 100.0, IMaxNoise(Is[0]));
    PrintTime(ci);
    for (i = 1; i < Value(QuietLevel); i++)
        {
        printf("     %11.6f   c=%d  m=%5.2f%%  n=%3d\n",
               PrintableError(IError(Is[i]), TS), ICrossingMethod(Is[i]),
               IMutationProb(Is[i]) / 100.0, IMaxNoise(Is[i]));
        }
    }
fflush(stdout);
}


/*-------------------------------------------------------------------*/


void PrintResult(TRAININGSET* TS,
                 INDIVIDUAL*  Is[],
                 int          Generation,
                 CLOCKINFO*   ci)
{
switch( Value(QuietLevel) )
    {
    case 0:
    case 1:
    case 2:  printf("\n%9.4f  (%i)  ", PrintableError(IError(Is[0]), TS),
                                       Generation); break;
    default: {
             printf("Distortion: %-9.4f\n", PrintableError(IError(Is[0]),TS));
             printf("Iterations: %-3i\n", Generation);
	     } break;
     }
PrintTime(ci);
}


/*-------------------------------------------------------------------*/


void PrintOperatingInfo(char* TSName, char* CBName)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Training Set:     %s\n", TSName);
    printf("Codebook:         %s\n", CBName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*====================  U T I L I T Y  =================================*/


void GenerateFileName(char* name, char* base, int i)
{
char number[10];

strcpy(name, base);

if (i > 999) printf("ERROR: too many codebooks\n"), exit(-1);
if (i < 100) strcat(name, "0");
if (i < 10)  strcat(name, "0");

sprintf(number, "%d", i);
strcat(name, number);
}


/*-------------------------------------------------------------------*/


int FreqToIndex(TRAININGSET* TS, int sum)
{
int i=0;
int l=0;

while (l<sum)
    {
    l += VectorFreq(TS, i);
    i++;
    }
return (i-1);
}


/*-------------------------------------------------------------------*/


YESNO ValidCrossingMethod(int c)
{
int i;

i = Value(CrossingMethodMask) >> (c-1);

return (i % 2 ? YES : NO);
}


/*==================  I N I T I A L I Z A T I O N  =====================*/


void Initialize(TRAININGSET*    TS, 
                INDIVIDUAL**    Is[],
                INDIVIDUAL**    NewIs[])
{                             
int i;
YESNO maskOK = NO;

if( Value(CodebookSize) >= BookSize(TS) )
    {
    printf("Codebook size (%d) > Training set size (%d).\n",
            Value(CodebookSize), BookSize(TS));
    exit(-1);
    }

for (i=0; i<=CrossingMethodCount; i++)
    if ( ValidCrossingMethod(i) ) maskOK = YES;
if ( maskOK == NO ) printf("Invalid crossing method mask.\n"), exit(-1);
    
*Is    = (INDIVIDUAL**) allocate(sizeof(INDIVIDUAL*) * Value(GenerationSize));
*NewIs = (INDIVIDUAL**) allocate(sizeof(INDIVIDUAL*) * Value(GenerationSize));

for (i=0; i<Value(GenerationSize); i++)
    {
    (*Is)[i]    = (INDIVIDUAL*) allocate( sizeof(INDIVIDUAL) );
    (*NewIs)[i] = (INDIVIDUAL*) allocate( sizeof(INDIVIDUAL) );
    CreateNewSolution(TS, &IGenes((*Is)[i]), Value(CodebookSize));
    CreateNewSolution(TS, &IGenes((*NewIs)[i]), Value(CodebookSize));
    }

initrandom(Value(RandomSeed));
}


/*-------------------------------------------------------------------*/


void InitializeVariables(int* SurvivorSetSize,
                         int* CrossingSetSize)
{
int s;

switch (Value(SelectionMethod))
    {
    case ROULETTEWHEEL:
        {
        *SurvivorSetSize = 1;
        *CrossingSetSize = Value(GenerationSize);
        break;
        }
    case ELITIST1:
        {
        s = 1;
        while ( (s*(s+1)/2) < Value(GenerationSize) ) s++;
        *SurvivorSetSize = *CrossingSetSize = s;
        break;
        }
    case ELITIST2:
        {
        s = 1;
        while ( (s*(s-1)/2) < Value(GenerationSize) ) s++;
        *SurvivorSetSize = 1;
        *CrossingSetSize = s;
        break;
        }
    }

if ( Value(NumberOfGenerations) < 1 )
    PrintInfo(), exit(-1);

SetPNNParameters(0,0,0,3,0,0,0,0,0);

}

                         
/*================  I N D I V I D U A L S  ==========================*/


void GenerateRandomVector(TRAININGSET* TS, SOLUTION* S, VECTORTYPE v)
{
/* Returns such vector v that v belongs to TS and v doesn't belong to S->CB */

int    i;
YESNO  ready;
int    t;
int    index;

do
    {
    ready=YES;

    t = irand(1, TotalFreq(TS));
    index = FreqToIndex(TS, t);

    for (i=0; i<Value(CodebookSize); i++)
        if (EqualVectors(Vector(&S->CB, i), Vector(TS, index), VectorSize(TS)))
            ready = NO;
    }
while (ready == NO);
CopyVector(Vector(TS, index), v, VectorSize(TS));
}


/*-------------------------------------------------------------------*/


void GenerateRandomSolution(TRAININGSET* TS, SOLUTION* S)
{
int i;
VECTORTYPE v;

v = CreateEmptyVector(VectorSize(TS));
/*
if (Value(PartitionBased))
    {
    for (i=0; i<BookSize(TS); i++)
        ChangePartition(TS, &(S->P), irand(0,Value(CodebookSize)-1), i);
    GenerateOptimalCodebook(TS, &(S->CB), &(S->P));
    }
else  */
    {
    for (i=0; i<Value(CodebookSize); i++)
        {
        GenerateRandomVector(TS, S, v);
        CopyVector(v, Vector(&S->CB, i), VectorSize(TS));
        VectorFreq(&S->CB, i) = 1;
        }
    GenerateOptimalPartitioning(TS, &S->CB, &S->P);
    }
FreeVector(v);
}


/*-------------------------------------------------------------------*/


void GenerateRandomIndividual(TRAININGSET* TS, INDIVIDUAL* I)
{
GenerateRandomSolution(TS, &IGenes(I));

do
    ICrossingMethod(I) = irand(1, CrossingMethodCount);
while (!(ValidCrossingMethod(ICrossingMethod(I))));
IMutationProb(I)   = irand(0, Value(MaxMutationProb));
IMaxNoise(I)       = irand(0, TS->MaxValue * Value(MaxNoise) / 100);

IError(I)          = SolutionError(TS, &IGenes(I), errorf);
}


/*--------------------------------------------------------------------*/


void AddNoiseToCodebook(CODEBOOK* CB, SASchedule* SAS)
{
int i;

if( !SASUseToCB(SAS) ) return;

for( i=0; i<BookSize(CB); i++ )
    RandomizeVectorBySA(SAS, &Node(CB, i), &Node(CB, i),
                        VectorSize(CB), CB->MaxValue);
}


/*-------------------------------------------------------------------*/


void SortIndividuals(TRAININGSET* TS, INDIVIDUAL* Is[])
  /* sorts individuals so that the best one will be Is[0] etc */
{
int         i,j;
INDIVIDUAL* I;

for (i=0; i<Value(GenerationSize); i++)
    for (j=0; j<i; j++)
        if ( IError(Is[i]) < IError(Is[j]) )
            {
            I = Is[i]; Is[i] = Is[j]; Is[j] = I;
            }             
}


/*-------------------------------------------------------------------*/


void CopyIndividual(INDIVIDUAL* I1, INDIVIDUAL* I2)
{
CopySolution(&IGenes(I1), &IGenes(I2));
ICrossingMethod(I2) = ICrossingMethod(I1);
IMutationProb(I2)   = IMutationProb(I1);
IMaxNoise(I2)       = IMaxNoise(I1);
IError(I2)          = IError(I1);
}


/*-------------------------------------------------------------------*/


void CrossParameters(INDIVIDUAL* I1, INDIVIDUAL* I2, INDIVIDUAL* Inew)
{
ICrossingMethod(Inew) = irand(0,1) == 0 ?
                        ICrossingMethod(I1) : ICrossingMethod(I2);
IMutationProb(Inew)   = irand(0,1) == 0 ?
                        IMutationProb(I1) : IMutationProb(I2);
IMaxNoise(Inew)       = irand(0,1) == 0 ?
                        IMaxNoise(I1) : IMaxNoise(I2);
}


/*-------------------------------------------------------------------*/


YESNO EqualPartitions(PARTITIONING* P1, PARTITIONING* P2)
{
int i;
int* inds;
int pc;

pc = P1->PartitionCount;

inds = allocate(pc * sizeof(int));
for (i=0; i<pc; i++) inds[i] = -1;

for (i=0; i<P1->TSsize; i++)
    {
    if (inds[Map(P1,i)] == -1) inds[Map(P1,i)] = Map(P2,i);
    else if (inds[Map(P1,i)] != Map(P2,i)) { deallocate(inds); return NO; }
    }
deallocate(inds);
return YES;
}


/*-------------------------------------------------------------------*/


YESNO EqualIndividuals(INDIVIDUAL* I1, INDIVIDUAL* I2)
{
/*
if ( ICrossingMethod(I1) != ICrossingMethod(I2) ) return NO;
if ( IMutationProb(I1)   != IMutationProb(I2)   ) return NO;
if ( IMaxNoise(I1)       != IMaxNoise(I2)       ) return NO;
if ( fabs(IError(I1) - IError(I2)) > EPSILON    ) return NO;
*/

if ( EqualPartitions(&IGenes(I1).P, &IGenes(I2).P) == NO ) return NO;

return YES;
}


/*======  C R O S S I N G   M E T H O D    U T I L S  ===============*/


void CopyRandomVectorFromCodebook(CODEBOOK* CB, BOOKNODE* node)
{
int n;

n=irand(0,BookSize(CB)-1);
CopyVector(Vector(CB, n), node->vector, VectorSize(CB));
node->freq=VectorFreq(CB,n);    
}


/*--------------------------------------------------------------------*/


void SortCodebookByCentroidDistance(TRAININGSET* TS,
                                    CODEBOOK*    CB)
  /* sorts codebook CB according to the distance
     from the training set centroid */
{
int        i,j;
llong*     CBdistances;
llong      t;
VECTORTYPE v;
VECTORTYPE centroid;

centroid = CreateEmptyVector(VectorSize(TS));

CBdistances = (llong*) allocate(BookSize(CB) * sizeof(llong));

/* calculate distances between code vectors and TS centroid */

CodebookCentroid(TS, centroid);

for (i=0; i<BookSize(CB); i++)
    CBdistances[i]=VectorDistance(Vector(CB,i),
                                  centroid,
                                  VectorSize(TS),
                                  MAXLLONG,
                                  errorf);
    
/* sort codebook according to the distances */

for (i=0; i<BookSize(CB); i++)
    for (j=0; j<i; j++)
        if (CBdistances[j]>CBdistances[i])
            {
            t=CBdistances[i]; CBdistances[i]=CBdistances[j]; CBdistances[j]=t;
            v=Vector(CB,i); Vector(CB,i)=Vector(CB,j); Vector(CB,j)=v;
            }

deallocate(CBdistances);
FreeVector(centroid);
}


/*============  I S   C R O S S I N G   U T I L S  ==================*/


int FindSecondNearest(CODEBOOK* TS, CODEBOOK* CB, int i, int tabu1, int tabu2)
/* Finds the nearest excluding candidates tabu1 and tabu2.
   The function can therefore find either 2nd or 3rd nearest. */
{
  int	j;
  int   MinIndex = 0;
  double factor;
  llong MinD = MAXLLONG;
  llong dist;

  for( j=0; j<BookSize(CB); j++ )
	 {
	 if(j==tabu1 || j==tabu2) continue;
     dist = VectorDist(Vector(CB,j), Vector(TS, i), VectorSize(CB));
     factor = ((double)VectorFreq(TS,i) * (double)VectorFreq(CB,j)) /
        ((double)VectorFreq(TS,i) + (double)VectorFreq(CB,j));
     dist = round( factor * (double)dist);
     if( dist<MinD ) { MinD=dist; MinIndex=j; }
     }
  return( MinIndex );
}


/*-------------------------------------------------------------------*/


static void GenerateSecondaryPartitioning(TRAININGSET*	TS,
										  CODEBOOK* 	CB,
										  PARTITIONING* P1,
										  PARTITIONING* P2)
{
  int  i,c;

  for( i=0; i<BookSize(TS); i++ )
	 {
     c = FindSecondNearest( TS, CB, i, -1, Map(P1,i) );
	  ChangePartition(TS, P2, c, i);
     }
}


/*-------------------------------------------------------------------*/


static void GenerateInitialPartitions(TRAININGSET*  TS,
                                      CODEBOOK*     CB,
                                      PARTITIONING* P1,
                                      PARTITIONING* P2)
{
  if( BookSize(TS) == BookSize(CB) )
	 {
	 PutAllInOwnPartition(TS, P1);
	 GenerateSecondaryPartitioning(TS, CB, P1, P2);
	 }
  else /* We have an initial codebook. */
     {
     GenerateOptimalPartitioning(TS, CB, P1);
     GenerateSecondaryPartitioning(TS, CB, P1, P2);
	}
}


/*--------------------------------------------------------------------*/


static void FindClusterToBeRemoved(CODEBOOK* CB,
								   llong	 RemovalCost[],
								   int* 	 cluster)
{
  llong MinDist=MAXLLONG;
  int i;

  for( i=0; i<BookSize(CB); i++ )
	 {
	 if( RemovalCost[i] < MinDist )
		{
		(*cluster) = i;
		MinDist = RemovalCost[i];
		}
	 }
  /*  if( Value(QuietLevel)>=6 ) printf("Cluster %i selected.\n",*cluster); */
}


/*-------------------------------------------------------------------*/


static void Init(int* Size, int Freq[], CODEBOOK* CB)
{
    int i;
    for( i=0; i<BookSize(CB); i++)
        Freq[i]=0;
    *Size=0;
}


/*-------------------------------------------------------------------*/


static void Clear(int* Size, int Elem[], int Freq[])
{
   int i;
   for( i=0; i<*Size; i++)
       Freq[Elem[i]]=0; 
   *Size=0;
}


/*-------------------------------------------------------------------*/


static void AddVector(llong Counter[],
                      int* vector, int dim, int ti, int freq)
{
    int k;
    for( k = 0; k < dim; k++ )
    {
        Counter[ti*(dim) + k] += vector[k]*freq;
    }
}


/*-------------------------------------------------------------------*/


static void CopVector(llong Counter[],
                      int* vector, int dim, int ti, int freq)
{
    int k;
    for( k = 0; k < dim; k++ )
    {
        Counter[ti*(dim) + k] = vector[k]*freq;
    }
}


/*-------------------------------------------------------------------*/


static void Add(int* Size, int Elem[], int Freq[], llong Counter[],
                int* vector, int dim, int ti, int freq)
{
    if(Freq[ti]==0)
    {
        Elem[*Size]=ti;
        *Size=*Size+1;
        CopVector(Counter, vector, dim, ti, freq);
        Freq[ti]=freq;
    }
    else
    {
        AddVector(Counter, vector, dim, ti, freq);
        Freq[ti]+=freq;
    }
}


/*-------------------------------------------------------------------*/


static void CalcCentroids(int ti, int Freq[], llong Counter[], 
                          CODEBOOK* Cents, BOOKNODE* bn)
{
    int k;

    for( k = 0; k<VectorSize(Cents); k++)
    {
        bn->vector[k] = round((double)Counter[ti *VectorSize(Cents) +k]
            /(double)Freq[ti]);
    }
}


/*-------------------------------------------------------------------*/


static void UpdatePrimaryPartition(TRAININGSET* TS, 
                                   PARTITIONING* P1, PARTITIONING* P2, 
                                   int* Size, int Elem[], int Freq[], 
                                   llong Counter[], int obsolete)
{
    int i, temp;
    Clear(Size, Elem, Freq);
    while ( UniqueVectors(P1, obsolete))
    {
        i = FirstVector(P1,obsolete);
        Add(Size, Elem, Freq, Counter, Vector(TS,i), VectorSize(TS), 
            Map(P2,i), VectorFreq(TS,i));
        temp = Map(P2,i);
        ChangePartition(TS, P2, Map(P1,i), i);
        ChangePartition(TS, P1, temp, i);
    }
}


/*-------------------------------------------------------------------*/


static void CalcPartitionCentroids(CODEBOOK* CB, 
                                   PARTITIONING* P1,
                                   int Size, 
                                   int Elem[])
{
    int j;
    for ( j = 0; j <Size; j++)
    {
        PartitionCentroid(P1, Elem[j], &Node(CB, Elem[j]));
    }
}


/*-------------------------------------------------------------------*/


static int MustBeUpdated(PARTITIONING* P1,
                         PARTITIONING* P2,
                         int Freq[],
                         int i,
                         int obsolete,
                         int UpdateStyle)
{
    int must = NO;
    switch( UpdateStyle)
      {
      case MINIMUM:
          {
          if( Map(P2, i) == obsolete)
              must=YES;
          break;
          }
      case STANDARD:
          {
          if( (Map(P2,i) == obsolete) || ( Freq[Map(P1,i)] && Freq[Map(P2,i)] ) )
              must=YES;
          break;
          }
      case EXTENSIVE:
          {
              if( (Map(P2,i) == obsolete) || Freq[Map(P1,i)] || Freq[Map(P2,i)] )
              must=YES;
          break;
          }
      default:
        {
            must=NO;
            break;
        }
      }
    return must;
}


/*-------------------------------------------------------------------*/


static void UpdateSecondaryPartition(TRAININGSET* TS,
									 CODEBOOK* CB,
									 PARTITIONING* P1,
									 PARTITIONING* P2,
                                     int Size,
                                     int Freq[],
                                     int obsolete,
                                     int UpdateStyle,
                                     int Quiet)
{
  int  i,ti;

  for( i=0; i<BookSize(TS); i++ )
  {
      if( MustBeUpdated(P1, P2, Freq, i, obsolete, UpdateStyle))
        {
           ti = FindSecondNearest(TS, CB, i, obsolete, Map(P1,i) );
	       ChangePartition(TS, P2, ti, i);
        }
  }
}


/*-------------------------------------------------------------------*/


static void RemoveObsolete(TRAININGSET* TS, 
                           CODEBOOK* CB, 
                           PARTITIONING* P1, 
                           PARTITIONING* P2, 
                           int obsolete)
{
    int last = BookSize(CB)-1;
    CopyNode(&Node(CB, last), &Node(CB, obsolete), VectorSize(CB));
    BookSize(CB)--; 
    JoinPartitions(TS, P1, obsolete, last);
    JoinPartitions(TS, P2, obsolete, last); 
}


/*-------------------------------------------------------------------*/


static llong CalcSubRemCost(CODEBOOK* CB, int Freq[], 
                            CODEBOOK* Cents, int ti)
    /* We merge the sub cluster (ti) and the cluster (ti) */
{
    double factor;
    llong dist;

    dist = VectorDist(Vector(Cents,ti), Vector(CB,ti), VectorSize(CB));
    factor = ((double)Freq[ti] * (double)VectorFreq(CB,ti)) /
        ((double)Freq[ti] + (double)VectorFreq(CB,ti));
    dist = round( factor * (double)dist);
    
    return (dist);
}


/*-------------------------------------------------------------------*/


static llong CalcRemCost(TRAININGSET* TS, 
                         CODEBOOK* CB, 
                         int i, 
                         int j)
    /* We remove the training vector (i) from the cluster (j) */
{

    llong dist;

    dist = VectorDist(Vector(TS,i), Vector(CB,j), VectorSize(CB)) * (llong)VectorFreq(TS,i);
 
    return (dist);
}


/*-------------------------------------------------------------------*/


static llong CalcMergeCost(TRAININGSET* TS, CODEBOOK* CB, int i, int j)
    /* We merge the training vector (i) and the code vector (j) */
{
    llong dist;
    double factor;

	dist = VectorDist(Vector(TS,i), Vector(CB,j), VectorSize(TS));
    factor=((double)VectorFreq(TS,i) * (double)VectorFreq(CB,j)) /
         ((double)VectorFreq(TS,i) + (double)VectorFreq(CB,j));
    dist = round( factor*(double)dist);
    return (dist);
}


/*-------------------------------------------------------------------*/


static void CalcRemovalCosts(TRAININGSET* TS, CODEBOOK* CB, 
                             PARTITIONING* P1, PARTITIONING* P2, 
                             int* Size, int Elem[], int Freq[], 
                             llong Counter[], CODEBOOK* Cents, 
                             llong RemCost[], int RemovalCalc)
{
   int i, j, l;

   Init(Size, Freq, CB);
   for( j=0; j <BookSize(CB); j++)
   {
     Clear(Size, Elem, Freq);
     RemCost[j]=0;
     for ( i = FirstVector(P1, j); ! EndOfPartition(i); i = NextVector(P1, i) )
     {
        Add(Size, Elem, Freq, Counter, Vector(TS,i), VectorSize(CB), 
            Map(P2,i), VectorFreq(TS,i));
        RemCost[j] -= CalcRemCost(TS, CB, i, j);
        if(RemovalCalc==SIMPLE)
        {
            RemCost[j] += CalcMergeCost(TS, CB, i, Map(P2,i));
        }
     }
     if(RemovalCalc==EXACT)
     {
     for( i = 0; i <*Size; i++)
        CalcCentroids(Elem[i], Freq, Counter, Cents, &Node(Cents,Elem[i]));
     for ( i = FirstVector(P1, j); ! EndOfPartition(i); i = NextVector(P1, i) )
     {  
        RemCost[j] += CalcRemCost(TS, Cents, i, Map(P2,i));  
     }
     for( l = 0; l <*Size; l++)
     {
        RemCost[j] += CalcSubRemCost(CB, Freq, Cents, Elem[l]);
     }
     }/* end RemovalCalc==EXACT */

   }
}


/*--------------------------------------------------------------------*/


static void RemoveEmptyPartitionsIS(TRAININGSET*  TS,
                                    CODEBOOK*     CB,
                                    PARTITIONING* P)
{
  int i,last;

  for( i=0; i<BookSize(CB); i++ )
     {
     if( VectorFreq(CB,i)==0  &&  BookSize(CB)>Value(CodebookSize) )
        {
        last = BookSize(CB)-1;
        CopyNode(&Node(CB,last), &Node(CB,i), VectorSize(CB));
        JoinPartitions(TS, P, i, last);
        BookSize(CB)--;
        i--;
        }
     }
}


/*-------------------------------------------------------------------*/


static void CreateDataStructures(CODEBOOK* CB, llong**  RemCost, int** Elem,
                               int** Freq, llong** Counter, CODEBOOK* Cents)
{
  *RemCost = (llong*) allocate( BookSize(CB) * sizeof(llong) );
  *Elem = (int*) allocate( BookSize(CB) * sizeof(int) );
  *Freq = (int*) allocate( BookSize(CB) * sizeof(int) );
  *Counter = (llong*) allocate( BookSize(CB) * VectorSize(CB) * sizeof(llong) );
  CreateNewCodebook(Cents, BookSize(CB), CB);
}


/*-------------------------------------------------------------------*/


static void FreeDataStructures(llong*  RemCost, 
                               int* Elem,
                               int* Freq, 
                               llong* Counter, 
                               CODEBOOK* Cents)
{
  deallocate(RemCost);
  deallocate(Elem);
  deallocate(Freq);
  deallocate(Counter);
  FreeCodebook(Cents);
}


/*-------------------------------------------------------------------*/


static void IterativeShrinking( TRAININGSET*   TS,
                                CODEBOOK*      CB,
                                PARTITIONING*  P1,
				                    PARTITIONING*  P2,
                                int            ResultCBSize,
                                int            RemovalCalc,
                                int            UpdateStyle,
                                int            Quiet)
{
  llong*    RemCost;
  int       Size;
  int*      Elem;
  int*      Freq;
  llong*    Counter;
  CODEBOOK  Cents;
  int       obsolete;
  
  CreateDataStructures(CB, &RemCost, &Elem, &Freq, &Counter, &Cents);
  RemoveEmptyPartitionsIS(TS, CB, P1);
  GenerateInitialPartitions(TS, CB, P1, P2);

  while( BookSize(CB)>ResultCBSize)
  {
    CalcRemovalCosts(TS, CB, P1, P2, &Size, Elem, Freq, Counter, &Cents, RemCost, RemovalCalc);
    FindClusterToBeRemoved(CB, RemCost, &obsolete);
    UpdatePrimaryPartition(TS, P1, P2, &Size, Elem, Freq, Counter, obsolete); 
    CalcPartitionCentroids(CB, P1, Size, Elem);
    UpdateSecondaryPartition(TS, CB, P1, P2, Size, Freq, obsolete, UpdateStyle, Quiet);
    RemoveObsolete(TS, CB, P1, P2, obsolete);
  }

  FreeDataStructures(RemCost, Elem, Freq, Counter, &Cents);
}


/*============  P N N   C R O S S I N G   U T I L S  =================*/


typedef struct { llong  VectorDist;
                 int    Vector;
                 YESNO  Recalculate;
               } NEARESTTYPE;


/*-------------------------------------------------------------------*/


llong MergeDistortion(CODEBOOK* CB, int p1, int p2)
{
llong  dist;
double factor;

if( VectorFreq(CB,p1)==0 || VectorFreq(CB,p2)==0 )
    dist = 0LL;
else
    {
    dist   = VectorDistance(Vector(CB,p1), Vector(CB,p2), VectorSize(CB),
                            MAXLLONG, errorf);
    factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
             ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
    dist   = (llong) ( factor * (double) dist + 0.5);
    }
return(dist);
}


/*-------------------------------------------------------------------*/


void FindNearestCluster(CODEBOOK* CB, int a, int* b, llong* dist)
{
llong distort;
int i;

(*dist)=MAXLLONG; (*b)=0;
for (i=0; i<BookSize(CB); i++)
    {
    if ( a!=i )
        {
        distort = MergeDistortion(CB, a, i);
        if ( distort < (*dist) )
            {
            (*dist) = distort;
            (*b)    = i;
            }
        }
    }
}


/*-------------------------------------------------------------------*/


void CalculateAllDistances(CODEBOOK* CB, NEARESTTYPE NN[])
{
int i;

for (i=0; i<BookSize(CB); i++)
    {
    FindNearestCluster(CB, i, &(NN[i].Vector), &(NN[i].VectorDist));
    }
}


/*-------------------------------------------------------------------*/


void FindMergedPair(CODEBOOK* CB, NEARESTTYPE NN[], int* a, int* b)
{
llong MinDist=MAXLLONG;
int i;

for (i=0; i<BookSize(CB); i++)
    {
    if( NN[i].VectorDist < MinDist )
        {
        (*a) = i;
        (*b) = NN[i].Vector;
        MinDist = NN[i].VectorDist;
        }
    }
if( (*a)>(*b) ) { i = *a; *a = *b; *b = i; }
}


/*-------------------------------------------------------------------*/


void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
int k;
int freqsum;

freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

/* Calculate mean values for centroid */
for (k=0; k<VectorSize(CB); k++)
    {
    if( freqsum > 0 )
        cent->vector[k] = (VECTORELEMENT) ((double)
                          (VectorFreq(CB, p1) * VectorScalar(CB, p1, k) +
                           VectorFreq(CB, p2) * VectorScalar(CB, p2, k))
                           / (double)freqsum  + 0.5);
    else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


/* ----------------------------------------------------------------- */


void RemoveLastVector(CODEBOOK* CB, int a, int b, NEARESTTYPE NN[])
{
int last = BookSize(CB)-1;
int i;

if (b != last)
    {
    CopyNode(&Node(CB,last), &Node(CB,b), VectorSize(CB));
    NN[b] = NN[last];
    for (i=0; i<BookSize(CB); i++)
        if (NN[i].Vector==last)  NN[i].Vector = b;
    }
}


/*--------------------------------------------------------------------*/


void RemoveEmptyPartitions(TRAININGSET* TS,
                           SOLUTION*    S)
{
int i, last;

for( i=0; i<BookSize(&S->CB); i++ )
    {
    if( VectorFreq(&S->CB,i) == 0 && BookSize(&S->CB) > Value(CodebookSize) )
        {
        last = BookSize(&S->CB)-1;
        CopyNode(&Node(&S->CB,last), &Node(&S->CB,i), VectorSize(TS));
        JoinPartitions(TS, &S->P, i, last);
        BookSize(&S->CB)--;
        i--;
        }
    }
}


/*--------------------------------------------------------------------*/


void MarkClustersForRecalculation(CODEBOOK*   CB,
                                  NEARESTTYPE NN[],
                                  int         a,
                                  int         b)
{
int i;

for (i=0; i<BookSize(CB); i++)
    NN[i].Recalculate = ( NN[i].Vector==a || NN[i].Vector==b ) ? YES : NO;
}


/*--------------------------------------------------------------------*/


void RecalculateDistances(CODEBOOK* CB, NEARESTTYPE NN[])
{
int  i;
int  nrec=0;

for (i=0; i<BookSize(CB); i++)
    {
    if (NN[i].Recalculate)
        {
        FindNearestCluster(CB, i, &(NN[i].Vector), &(NN[i].VectorDist));
        NN[i].Recalculate = NO;
        nrec++;
        }
    }
}


/*-------------------------------------------------------------------*/


void MergeVectors(TRAININGSET* TS,
                  SOLUTION*    S,
                  NEARESTTYPE  NN[],
                  int          a,
                  int          b)
{
MarkClustersForRecalculation(&S->CB, NN, a, b);
CentroidOfTwoVectors(&S->CB, &Node(&S->CB, a), a, b);
JoinPartitions(TS, &S->P, a, b);
RemoveLastVector(&S->CB, a, b, NN);
BookSize(&S->CB)--;
}


/*-------------------------------------------------------------------*/


void PNNFast(TRAININGSET* TS,
             SOLUTION*    S,
             int          ResultCBSize)
{
NEARESTTYPE* NN;
int          a, b;

RemoveEmptyPartitions(TS, S);
NN = (NEARESTTYPE*) allocate(BookSize(&S->CB) * sizeof(NEARESTTYPE));
CalculateAllDistances(&S->CB, NN);
while( BookSize(&S->CB) > ResultCBSize )
    {
    FindMergedPair(&S->CB, NN, &a, &b);
    MergeVectors(TS, S, NN, a, b);
    RecalculateDistances(&S->CB, NN);
    }
deallocate(NN);
}


/*===============  C R O S S I N G   M E T H O D S  =================*/


void CrossRandom(TRAININGSET* TS,
                 SOLUTION*    S1,
                 SOLUTION*    S2,
                 SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using RANDOM CROSSOVER method.
   Generating method is to select random vectors from both codebooks and
   check whether they already exist in the new one. */
{
int     i,j;
YESNO   ready;

for (i=0; i<BookSize(&Snew->CB); i++)
    {
    do
        {
        ready=YES;
        CopyRandomVectorFromCodebook( (i%2==0)? &S1->CB : &S2->CB,
                                      &Node(&Snew->CB,i));
        for (j=0; j<i; j++)
            if (EqualVectors(Vector(&Snew->CB, i),Vector(&Snew->CB, j),
                             VectorSize(TS)))
                ready=NO;
        }
    while (ready==NO);
    }        

Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossCentroidDistance(TRAININGSET* TS,
                           SOLUTION*    S1,
                           SOLUTION*    S2,
                           SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using ORDERING BY CENTROID DISTANCE method.
   Generating method is to order both codebooks according to the distance
   from centroid and then "swap tails". Finally duplicates are replaced with
   random vectors from codebooks. Function assumes that S1 and S2 are
   free of duplicates. */
{
int       i, j;
CODEBOOK  *cb1, *cb2;
    
SortCodebookByCentroidDistance(TS, &S1->CB);
SortCodebookByCentroidDistance(TS, &S2->CB);
        
/* create new from S1->CB and S2-CB */

if (irand(0,1)==0)
    cb1 = &S1->CB, cb2 = &S2->CB;
else
    cb1 = &S2->CB, cb2 = &S1->CB;

for (i=0; i<Value(CodebookSize)/2; i++)
    CopyVector(Vector(cb1,i), Vector(&Snew->CB,i), VectorSize(TS));
        
for (i=Value(CodebookSize)/2; i<Value(CodebookSize); i++)
    CopyVector(Vector(cb2,i), Vector(&Snew->CB,i), VectorSize(TS));
    
/* replace the duplicates */

for (i=Value(CodebookSize)/2; i<Value(CodebookSize); i++)
    for (j=0; j<Value(CodebookSize)/2; j++)
        {
            while (EqualVectors(Vector(&Snew->CB,i),Vector(&Snew->CB,j),
                                VectorSize(TS)))
            CopyRandomVectorFromCodebook( (irand(0,1)==0)? cb1 : cb2,
                                          &Node(&Snew->CB,i));
        }    
Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossLargestPartitions(TRAININGSET* TS,
                            SOLUTION*    S1,
                            SOLUTION*    S2,
                            SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using LARGEST PARTITIONS method.
   Generating method is to always select the most popular codevector
   and remove the effect of  codevectors already selected. This seems
   to be rather slow. */
{
int   i;
int   *CB1pop, *CB2pop;           /* popularities */
int   *TR1map, *TR2map;           /* mappings */
int   bestind = 0;
int   bestpop;
int   CBstate;                    /* number of codevector in process */
llong e;

CB1pop = (int*) allocate(Value(CodebookSize) * sizeof(int));
CB2pop = (int*) allocate(Value(CodebookSize) * sizeof(int));
TR1map = (int*) allocate(BookSize(TS) * sizeof(int));
TR2map = (int*) allocate(BookSize(TS) * sizeof(int));

for (i=0; i<Value(CodebookSize); i++)
    CB1pop[i] = CB2pop[i] = 0;
    
for (i=0; i<BookSize(TS); i++)
    {
    bestind = FindNearestVector(&Node(TS,i), 
                                &S1->CB,
                                &e, 
                                Value(CodebookSize)-1,
                                errorf);
    TR1map[i]=bestind; CB1pop[bestind]++;
    bestind = FindNearestVector(&Node(TS,i), 
                                &S2->CB,
                                &e, 
                                Value(CodebookSize)-1,
                                errorf);
    TR2map[i]=bestind; CB2pop[bestind]++;
    }
CBstate=0;

do
    {
    /* find the most popular codevector */

bestpop=-1;
    for (i=0; i<Value(CodebookSize); i++)
        {
        if (CB1pop[i]>bestpop)
            bestpop=CB1pop[i], bestind=i;
        if (CB2pop[i]>bestpop)    
            bestpop=CB2pop[i], bestind=i+Value(CodebookSize);
        }

    /* copy the vector to the new codebook */               

    if (bestind < Value(CodebookSize))
        CopyVector(Vector(&S1->CB, bestind),
                   Vector(&Snew->CB, CBstate),
                   VectorSize(&Snew->CB));
    else
        CopyVector(Vector(&S2->CB, bestind-Value(CodebookSize)),
                   Vector(&Snew->CB, CBstate),
                   VectorSize(TS));

    /* remove the effect of training vectors mapping to Snew->CB[i] */

    for (i=0; i<BookSize(TS); i++)
        {
        if (bestind<Value(CodebookSize))
            {
            if (TR1map[i] == bestind)
                CB2pop[TR2map[i]]--;
            }
        else
            {
            if (TR2map[i] == bestind - Value(CodebookSize))
                CB1pop[TR1map[i]]--;
            }
        }
        
    /* duplicate checking */

    for (i=0; i<CBstate; i++)
        {
        if (EqualVectors(Vector(&Snew->CB, i),
                         Vector(&Snew->CB, CBstate),
                         VectorSize(TS)))
            {
            CBstate--;
            break;
            }
        }
    CBstate++;        

    /* set the popularity of selected vector to -1 so that it won't be
       chosen again */

    if (bestind < Value(CodebookSize))
        CB1pop[bestind] = -1;
    else
        CB2pop[bestind - Value(CodebookSize)] = -1;
        
    }
while (CBstate < Value(CodebookSize));
    
deallocate(CB1pop);
deallocate(CB2pop);
deallocate(TR1map);
deallocate(TR2map);

Snew->optimality = OPT_NONE;
} 


/*--------------------------------------------------------------------*/


void CrossPairwiseRandom(TRAININGSET* TS,
                         SOLUTION*    S1,
                         SOLUTION*    S2,
                         SOLUTION*    Snew)
  /* Generates Snew from S1 and S2
     using RANDOM PAIRWISE CROSSOVER method. First a (non-optimal)
     bipartition assignment is made and then the new codebook is
     generated by selecting randomly one of assigned vectors in each
     place. */
{
int   i, j;
int   *unsorted; /* such i's that S2->CB[i] hasn't been chosen yet  */
int   *sorted;   /* S1->CB[i] will be close to S2->CB[sorted[i]]. */
llong dist, best;
int   bestind;

unsorted = (int*) allocate(Value(CodebookSize) * sizeof(int));
sorted   = (int*) allocate(Value(CodebookSize) * sizeof(int));

for (i=0; i<Value(CodebookSize); i++)
    unsorted[i]=i;
    
for (i=0; i<Value(CodebookSize); i++)
    {
    /* first find the closest match of S1->CB[i] in S2->CB that
       hasn't been chosen yet */
    best = MAXLLONG;
    bestind = 0;
    for (j=0; j<Value(CodebookSize)-i; j++)
        {
        dist = VectorDistance(Vector(&S1->CB, i),
                              Vector(&S2->CB, unsorted[j]),
                              VectorSize(TS), best, errorf);
        if (dist < best)
            best = dist, bestind = j;
        }
    /* the closest match is S2->CB[unsorted[bestind]] - save to sorted and
       remove from unsorted */
    sorted[i] = unsorted[bestind];
    unsorted[bestind] = unsorted[Value(CodebookSize)-i-1];
    }
    
/* do N point crossover */

for (i=0; i<Value(CodebookSize); i++)
    {
    if (irand(0,1)==0)
        CopyVector(Vector(&S1->CB,i), Vector(&Snew->CB,i), VectorSize(TS));
    else
        CopyVector(Vector(&S2->CB,sorted[i]), Vector(&Snew->CB,i),
                   VectorSize(TS));
    }        

/* finally, remove duplicates */

for (i=0; i<Value(CodebookSize); i++)
    for (j=0; j<i; j++)
        {
        while (EqualVectors(Vector(&Snew->CB,i),Vector(&Snew->CB,j),
                            VectorSize(TS)))
        CopyRandomVectorFromCodebook( (irand(0,1)==0)? &S1->CB : &S2->CB,
                                      &Node(&Snew->CB,i));
        }    

deallocate(unsorted);
deallocate(sorted);

Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossPairwiseOnePoint(TRAININGSET* TS,
                           SOLUTION*    S1,
                           SOLUTION*    S2,
                           SOLUTION*    Snew)
  /* Generates Snew from S1 and S2 using PAIRWISE ONE POINT CROSSOVER method.
     Bipartition assignment similar to CrossSolutions[3] is made
     but crossing is made by swapping tails (see CrossSolutions[1]). */
{
int   i,j;
int   *unsorted; /* such i's that S2->CB[i] hasn't been chosen yet  */
int   *sorted;   /* CB[old1][i] will be close to S2->CB[sorted[i]]. */
llong dist,best;
int   bestind;
YESNO firstfirst; /* do we take vectors 0...CBsize/2 from S1->CB? */

unsorted = (int*) allocate(Value(CodebookSize) * sizeof(int));
sorted   = (int*) allocate(Value(CodebookSize) * sizeof(int));

for (i=0; i<Value(CodebookSize); i++)
    unsorted[i]=i;
    
for (i=0; i<Value(CodebookSize); i++)
    {
    /* first find the closest match for S1->CB[i] in S2->CB that
       hasn't been chosen yet */
    best = MAXLLONG; bestind = 0;
    for (j=0; j<Value(CodebookSize)-i; j++)
        {
        dist=VectorDistance(Vector(&S1->CB, i),
                            Vector(&S2->CB, unsorted[j]),
                            VectorSize(TS), best, errorf);
        if (dist < best)
            best = dist, bestind = j;
        }
    /* the closest match is S2->CB[unsorted[bestind]] - save to sorted and
       remove from unsorted */
    sorted[i] = unsorted[bestind];
    unsorted[bestind] = unsorted[Value(CodebookSize)-i-1];
    }

/* create new codebook from S1->CB and S2->CB */

firstfirst = (irand(0,1)==0);

for (i=0; i<Value(CodebookSize)/2; i++)
    if (firstfirst)
        CopyVector(Vector(&S1->CB,i), Vector(&Snew->CB,i), VectorSize(TS));
    else
        CopyVector(Vector(&S2->CB,sorted[i]), Vector(&Snew->CB,i),
                   VectorSize(TS));
        
for (i=Value(CodebookSize)/2; i<Value(CodebookSize); i++)
    if (firstfirst)
        CopyVector(Vector(&S2->CB,sorted[i]), Vector(&Snew->CB,i),
                   VectorSize(TS));
    else
        CopyVector(Vector(&S1->CB,i), Vector(&Snew->CB,i),
                   VectorSize(TS));

/* replace the duplicates */

for (i=Value(CodebookSize)/2; i<Value(CodebookSize); i++)
    for (j=0; j<Value(CodebookSize)/2; j++)
        {
        while (EqualVectors(Vector(&Snew->CB,i),Vector(&Snew->CB,j),
                            VectorSize(TS)))
            CopyRandomVectorFromCodebook( (irand(0,1)==0)? &S1->CB : &S2->CB,
                                           &Node(&Snew->CB,i));
        }    

deallocate(unsorted);
deallocate(sorted);

Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossPNN1(TRAININGSET* TS,
               SOLUTION*    S1,
               SOLUTION*    S2,
               SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using the oldest PNN crossing method.
   A training set (of size 2M) is created from the input codebooks and a
   codebook (of size M) is generated for this training set. This crossing
   method first appeared in-built in version 0.60. */
{
int         i;
TRAININGSET TSPNN;
CODEBOOK    CBPNN;

CreateNewCodebook(&TSPNN, 2 * Value(CodebookSize), TS);
CreateNewCodebook(&CBPNN, 2 * Value(CodebookSize), TS);

for (i=0; i<Value(CodebookSize); i++)
    {
    CopyNode(&Node(&S1->CB,i), &Node(&TSPNN,i), VectorSize(TS));
    VectorFreq(&TSPNN,i) = 1;
    CopyNode(&Node(&S2->CB,i), &Node(&TSPNN,i+Value(CodebookSize)),
             VectorSize(TS));
    VectorFreq(&TSPNN,i+Value(CodebookSize)) = 1;
    }

CopyCodebook(&TSPNN, &CBPNN);
PairwiseNearestNeighbour(&TSPNN, &CBPNN, NULL, NULL, Value(CodebookSize));

CopyCodebook(&CBPNN, &Snew->CB);
FreeCodebook(&TSPNN);
FreeCodebook(&CBPNN);

Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossPNN2(TRAININGSET* TS,
               SOLUTION*    S1,
               SOLUTION*    S2,
               SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using the first PNN crossing method
   implemented in CBgenet. Input codebooks are combined as a codebook
   of size 2M. From this, a codebook of size M is created by PNN
   algorithm. Training set is not altered. */

{
int      i;
CODEBOOK CB;

CreateNewCodebook(&CB, 2*Value(CodebookSize), TS);

for (i=0; i<Value(CodebookSize); i++)
    {
    CopyNode(&Node(&S1->CB,i), &Node(&CB,i), VectorSize(TS));
    CopyNode(&Node(&S2->CB,i), &Node(&CB,i+Value(CodebookSize)),
             VectorSize(TS));
    }

PairwiseNearestNeighbour(TS, &CB, NULL, NULL, Value(CodebookSize));
CopyCodebook(&CB, &Snew->CB);
FreeCodebook(&CB);

Snew->optimality = OPT_NONE;
}


/*--------------------------------------------------------------------*/


void CrossPNN3(TRAININGSET* TS,
               SOLUTION*    S1,
               SOLUTION*    S2,
               SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using the latest PNN crossing method.
   The effect is equal to CrossPNN2, but this is a much faster
   implementation.*/
{
SOLUTION S;
llong    d1, d2;
int      i, p1, p2;

CreateNewSolution(TS, &S, 2*Value(CodebookSize));

/* Combine codebooks */
for (i=0; i<Value(CodebookSize); i++)
    {
    CopyNode(&Node(&S1->CB,i), &Node(&S.CB,i), VectorSize(TS));
    CopyNode(&Node(&S2->CB,i), &Node(&S.CB,i+Value(CodebookSize)),
             VectorSize(TS));
    }

/* Combine partitions: P3 <- Best partitions from P1 and P2 */
CopyPartitioning(&S1->P, &S.P);
PartitionCount(&S.P) = 2*Value(CodebookSize);
for(i=0; i<BookSize(TS); i++)
    {
    p1 = Map(&S1->P, i);
    p2 = Map(&S2->P, i);
    d1 = VectorDistance(Vector(TS,i), Vector(&S1->CB,p1), VectorSize(TS),
                        MAXLLONG, errorf);
    d2 = VectorDistance(Vector(TS,i), Vector(&S2->CB,p2), VectorSize(TS),
                        MAXLLONG, errorf);
    if (d2 < d1) ChangePartition(TS, &S.P, Value(CodebookSize)+p2, i);
    }

/* Calculate optimal codebook (2M) for combined partitioning */
ChangeOptimality(TS, &S, OPT_CB, errorf);

/* Reduce codebook size from 2M -> M using PNN algorithm */
PNNFast(TS, &S, Value(CodebookSize));
CopySolution(&S, Snew);
FreeSolution(&S);
}


/*--------------------------------------------------------------------*/


static void CrossIS(TRAININGSET* TS,
                    SOLUTION*    S1,
                    SOLUTION*    S2,
                    SOLUTION*    Snew)
/* Generates Snew from S1 and S2 using IS crossing method.  */
{
  llong  d1, d2;
  int    i, p1, p2;
  int    booksize = BookSize(&S1->CB);
  int    vsize = VectorSize(&S1->CB);
  PARTITIONING  P2;
  CreateNewPartitioning(&P2, TS, 2*Value(CodebookSize));

  /*  if( Value(QuietLevel)>=5 )
      { printf("--> Combining codebooks by IS."); } */

  /* Combine partitions: P3 <- Best partitions from P1 and P2 */
  CopyPartitioning(&S1->P, &Snew->P);
  /* Combine codebooks: CB3 <- CB1 + CB2 */
  IncreaseSolutionSize(Snew, 2*Value(CodebookSize));
  for(i=0; i<booksize; i++)
     CopyNode( &Node(&S1->CB,i), &Node(&Snew->CB,i), vsize );
  for(i=0; i<booksize; i++)
     CopyNode( &Node(&S2->CB,i), &Node(&Snew->CB,booksize+i), vsize );

  /*  if( Value(QuietLevel)>=5 )
      { printf("--> Combining partitionings."); } */

  /*   PartitionCount(&Snew->P) = 2*Value(CodebookSize);  */
  for(i=0; i<BookSize(TS); i++)
    {
    p1 = Map(&S1->P, i);
    p2 = Map(&S2->P, i);
    d1 = VectorDist(Vector(TS,i), Vector(&S1->CB,p1), vsize);
    d2 = VectorDist(Vector(TS,i), Vector(&S2->CB,p2), vsize);
    if(d2<d1) ChangePartition(TS, &Snew->P, booksize+p2, i);
    }

  /* Calculate optimal codebook (2M) for combined partitioning */
  GenerateOptimalCodebook(TS, &Snew->CB, &Snew->P);
  Snew->optimality = OPT_CB;

  /* Reduce codebook size from 2M -> M using IS algorithm */
  /*  if( Value(QuietLevel)>=5 )
      { printf("--> Reducing codebook by IS."); } */
  IterativeShrinking(TS, &Snew->CB, &Snew->P, &P2, Value(CodebookSize),
      Value(RemovalCostCalculation), Value(Update), Value(QuietLevel));

  /*  if( Value(QuietLevel)>=5 )
      { printf("--> Crossing completed."); } */
  FreePartitioning(&P2);
}


/*===============  G E N E T I C   O P E R A T I O N S  =============*/


void MutateSolution(TRAININGSET* TS, SOLUTION* S, int mutations)
{
int i;
VECTORTYPE v;

v = CreateEmptyVector(VectorSize(TS));
for (i=0; i<BookSize(&S->CB); i++)
    if (irand(0, 9999) < mutations)
        {
        GenerateRandomVector(TS, S, v);
        ChangeCodeVectorInSolution(TS, S, i, v, errorf,
                                   S->optimality == OPT_PA ||
                                   S->optimality == OPT_BOTH);
        }
FreeVector(v);
}


/*--------------------------------------------------------------------*/


void AddNoiseToSolution(TRAININGSET* TS, SOLUTION* S, int maxtemp)
{
SASchedule SAS;
int i;

InitializeSASchedule(0, 0, 0, (double) maxtemp, 1, 0, &SAS);

for (i=0; i<BookSize(&S->CB); i++)
    {
    RandomizeVectorBySA(&SAS, &Node(&S->CB, i), &Node(&S->CB, i),
                        VectorSize(TS), TS->MaxValue);
    }
}


/*--------------------------------------------------------------------*/


void MutateParameters(TRAININGSET* TS, INDIVIDUAL* I)
{
int maxn;

if (irand(0, 9999) < Value(ParamMutationProb) )
    do
	ICrossingMethod(I) = irand(1, CrossingMethodCount);
    while (!(ValidCrossingMethod(ICrossingMethod(I))));
if (irand(0, 9999) < Value(ParamMutationProb) )
    {
    IMutationProb(I) += (int) (dgauss() * Value(MaxMutationProb) / 22.0),
    IMutationProb(I) = LimitN(IMutationProb(I), 0, Value(MaxMutationProb));
    }
if (irand(0, 9999) < Value(ParamMutationProb) )
    {
    maxn = TS->MaxValue * Value(MaxNoise) / 100;
    IMaxNoise(I) += (int) (dgauss() * maxn / 22.0);
    IMaxNoise(I) = LimitN(IMaxNoise(I), 0, maxn);
    }
}


/*--------------------------------------------------------------------*/


void MutateAndAddNoise(TRAININGSET* TS, INDIVIDUAL* I)
{
MutateSolution(TS, &IGenes(I), IMutationProb(I));
if ( IMaxNoise(I) > 0 )
    {
    AddNoiseToSolution(TS, &IGenes(I), IMaxNoise(I));
    IGenes(I).optimality = OPT_NONE;
    }

/* parameters are mutated before crossing */
}


/*--------------------------------------------------------------------*/


void MutateDuplicates(TRAININGSET* TS, INDIVIDUAL* Is[])
{
int i;
VECTORTYPE v;
YESNO changed = YES;
int r;
int low;

v = CreateEmptyVector(VectorSize(TS));
while ( changed )
    {
    changed = NO;
    for (i=Value(GenerationSize)-1; i>0; i--)
        {
        low = i-1;
        while ( low >= 0 && IError(Is[i]) - IError(Is[low]) < EPSILON ) low--;
        low++;
        while ( low < i && !(EqualIndividuals(Is[i], Is[low])) ) low++;
        if ( low != i )
            {
            changed = YES;
            r = irand(0, Value(CodebookSize) - 1);
            GenerateRandomVector(TS, &IGenes(Is[i]), v);
            ChangeCodeVectorInSolution(TS, &IGenes(Is[i]), r, v, errorf,
                                       IGenes(Is[i]).optimality == OPT_PA ||
                                       IGenes(Is[i]).optimality == OPT_BOTH);
            IError(Is[i]) = SolutionError(TS, &IGenes(Is[i]), errorf);
            }
        }
    if (changed) SortIndividuals(TS, Is);
    }
FreeVector(v);
}


/*--------------------------------------------------------------------*/


int PickRandomSolution(INDIVIDUAL* Is[], int CrossingSetSize)

/* Selects a random solution by roulette wheel. Only CrossingSetSize
   first solutions are to be considered                              */
{
double total = 0;
double r;
int    i;

for (i=0; i<CrossingSetSize; i++)
    total += 1.0 / (1.0 + IError(Is[i]));

r = drand() * total;
total=0;

for (i=0; i<CrossingSetSize; i++)
    {
    total += 1 / (1 + IError(Is[i]));
    if (total >= r) return i;
    }

printf("Unexpected ERROR in PickRandomSolution. Darn.\n");
exit(-1);
return 0;
}


/*--------------------------------------------------------------------*/


void DetermineParentsByRouletteWheel(int*        pairs,
                                     INDIVIDUAL* Is[],
                                     int         CrossingSetSize,
                                     int         ParentCount)
/* ParentCount indicates the number of pairs to be generated */
{
int i;

for (i=0; i<ParentCount; i++)
    {
    pairs[2*i] = PickRandomSolution(Is, CrossingSetSize);
    while ( (pairs[2*i+1] = PickRandomSolution(Is, CrossingSetSize))
                          == pairs[2*i] );
    }
}

    
/*--------------------------------------------------------------------*/


void DetermineParentsByElitism(int* pairs,
                               int  CrossingSetSize)
{
int i,j,k;

k = 0;
for (i=0; i<CrossingSetSize; i++)
    for (j=0; j<i; j++)
        {
        pairs[k] = i;
        pairs[k+1] = j;
        k += 2;
        }
}


/*--------------------------------------------------------------------*/


void IterateGLAForSolutions(TRAININGSET* TS,
                            INDIVIDUAL*  Is[],
                            int          firstnew)
{
/* also updates errors and changes optimality to OPT_PA */
int i;

if (Value(GLAIterations) == 0)
    {
    for (i=firstnew; i<Value(GenerationSize); i++)
        {
        ChangeOptimality(TS, &IGenes(Is[i]), OPT_PA, errorf);
        IError(Is[i]) = SolutionError(TS, &IGenes(Is[i]), errorf);
        }
    return;
    }

for (i = firstnew; i<Value(GenerationSize); i++)
    {
    IterateGLAForSolution(TS, &IGenes(Is[i]), Value(GLAIterations), errorf);
    IError(Is[i]) = SolutionError(TS, &IGenes(Is[i]), errorf);
    }
}


/*--------------------------------------------------------------------*/

    
void CrossIndividuals(TRAININGSET* TS,
                      INDIVIDUAL*  I1,
                      INDIVIDUAL*  I2,
                      INDIVIDUAL*  Inew)
{
SOLUTION* S1   = &IGenes(I1);
SOLUTION* S2   = &IGenes(I2);
SOLUTION* Snew = &IGenes(Inew);

CrossParameters(I1, I2, Inew);
MutateParameters(TS, Inew);

switch ( ICrossingMethod(Inew) )
    {
    case 1: CrossRandom(TS, S1, S2, Snew);            break;
    case 2: CrossCentroidDistance(TS, S1, S2, Snew);  break;
    case 3: CrossLargestPartitions(TS, S1, S2, Snew); break;
    case 4: CrossPairwiseRandom(TS, S1, S2, Snew);    break;
    case 5: CrossPairwiseOnePoint(TS, S1, S2, Snew);  break;
    case 6: CrossPNN3(TS, S1, S2, Snew);              break;
    case 7: CrossIS(TS, S1, S2, Snew);                break;
    default: printf("ERROR: invalid crossing method %d\n",
                    ICrossingMethod(Inew)); exit(1);   break;
    }
}

    
/*--------------------------------------------------------------------*/


void GenerateNewGeneration(TRAININGSET* TS,
                           INDIVIDUAL*  Is[],
                           INDIVIDUAL*  NewIs[],
                           int          SurvivorSetSize,
                           int          CrossingSetSize)
{
int   i;
int*  pairs;      /* contains pairs to be crossed - pairs[0] & pairs[1] etc */

pairs = allocate( sizeof(int) * 2 * Value(GenerationSize));

for (i=0; i < SurvivorSetSize; i++)
    CopyIndividual(Is[i], NewIs[i]);

switch (Value(SelectionMethod))
    {
    case ROULETTEWHEEL:
        DetermineParentsByRouletteWheel(pairs, Is, CrossingSetSize,
                                     Value(GenerationSize) - SurvivorSetSize);
        break;
    case ELITIST1:
    case ELITIST2:
        DetermineParentsByElitism(pairs, CrossingSetSize);
        break;
    }
for (i=0; i<Value(GenerationSize) - SurvivorSetSize; i++)
    CrossIndividuals(TS, Is[pairs[2*i]], Is[pairs[2*i+1]],
                     NewIs[i+SurvivorSetSize]);

for (i=SurvivorSetSize; i< Value(GenerationSize); i++)
    {
    MutateAndAddNoise(TS, NewIs[i]);
    }
        
IterateGLAForSolutions(TS, NewIs, SurvivorSetSize);
SortIndividuals(TS, NewIs);
if (Value(RemoveDuplicates)) MutateDuplicates(TS, NewIs);
deallocate(pairs);
}


/*=====================  S H U T D O W N  ============================*/


void ShutDown(TRAININGSET*   TS, 
              INDIVIDUAL*    Is[],
              INDIVIDUAL*    NewIs[])
{
int i;

FreeCodebook(TS);

for (i=0; i<Value(GenerationSize); i++)
    {
    FreeSolution(&IGenes(Is[i]));
    FreeSolution(&IGenes(NewIs[i]));
    deallocate(Is[i]);
    deallocate(NewIs[i]);
    }
    
deallocate(Is);
deallocate(NewIs);
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
char           InFileName[80], OutFileName[80];
YESNO          EscPressed = NO;
int            i, Generation = 1;
TRAININGSET    TS;
INDIVIDUAL     **Is, **NewIs, **TempIP;
CLOCKINFO      ci = {0,0};
int            SurvivorSetSize, CrossingSetSize;

ParameterInfo  paraminfo[2] =  { { InFileName,  FormatNameTS, 0, INFILE   },
                                 { OutFileName, "",           0, DONTCARE } };

ParseParameters( argc, argv, 2, paraminfo );
InitializeVariables(&SurvivorSetSize, &CrossingSetSize);
ReadTrainingSet(InFileName, &TS);
Initialize(&TS, &Is, &NewIs);

/* first generation - randomly generated */
for (i=0; i<Value(GenerationSize); i++)
    GenerateRandomIndividual(&TS, Is[i]);
SortIndividuals(&TS, Is);

/* iteration */
while (Generation <= Value(NumberOfGenerations) && !(EscPressed))
    {
    UpdateClock(&ci);
    PrintProgress(&TS, Is, Generation, &ci);
    Generation++;
    GenerateNewGeneration(&TS, Is, NewIs, SurvivorSetSize, CrossingSetSize);
    /* swap old individuals and new individuals */
    TempIP = NewIs; NewIs = Is; Is = TempIP;
    #if defined(__MSDOS__)
    if (kbhit())
        EscPressed = (getch() == 27);
    #endif
    }

Generation--;
PrintResult(&TS, Is, Generation, &ci);

CheckFileName(OutFileName, FormatNameCB);
WriteCodebook(OutFileName, &(IGenes(Is[0]).CB), YES);

ShutDown(&TS, Is, NewIs);
checkmemory();
return 0;
}
